import sys
from io import StringIO

class Capture(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio
        sys.stdout = self._stdout



import uuid
import sys
import types
import pickle
import base64
import threading
import queue
import logging
import time

import pysparkrpc
from pysparkrpc.server.capture import Capture
from pysparkrpc.server.logger import logger, configure_logging

logger = logging.getLogger()

import pyspark
import pyspark.sql.functions
import cloudpickle

from flask import Flask, request, jsonify
from flask.logging import default_handler
app = Flask(__name__)

OBJECTS = {}
RESPONSE_CACHE = {}

REQ_QUEUE = queue.Queue()
RESP_QUEUE = queue.Queue()

def retrieve_object(obj):
    global OBJECTS

    if type(obj) == dict and '_PROXY_ID' in obj:
        id = obj['_PROXY_ID']
        logger.info('Retrieving object id: %s' % id)

        return OBJECTS[id]
    else:
        return obj

def parse_request_args(function, request_args, request_kwargs):
    global OBJECTS

    args = []
    kwargs = {}

    if function in pysparkrpc.PICKLE_FUNCS:
        args = cloudpickle.loads(base64.b64decode(request_args))
        kwargs = cloudpickle.loads(base64.b64decode(request_kwargs))
    else:
        for a in request_args:
            arg_type = type(a)

            if arg_type == dict:
                if '_CLOUDPICKLE' in a:
                    obj = cloudpickle.loads(base64.b64decode(a['_CLOUDPICKLE']))
                    args.append(obj)
                else:
                    args.append(retrieve_object(a))
            # pyspark objects can sometimes be in lists so we need to
            # check the list and lookup any objects
            elif arg_type == list or arg_type == tuple:
                processed_list = []

                for x in a:
                    type_x = type(x)

                    if type_x == list or type_x == tuple:
                        processed_sub_list = []

                        for sub_x in x:
                            processed_sub_list.append(retrieve_object(sub_x))

                        processed_list.append(processed_sub_list)
                    else:
                        processed_list.append(retrieve_object(x))

                args.append(processed_list)
            # spark strictly typechecks some arguments expecting strings but
            # decoding json will turn strings into unicode objects
            elif arg_type == str:
                args.append(str(a))
            else:
                args.append(a)

        for k in request_kwargs:
            v = request_kwargs[k]

            kwargs[k] = retrieve_object(v)

    return args, kwargs

# checks if any objects are created via a function call
# and returns the corresponding object
#
# pandas: pickles and returns a pandas dataframe
# pyspark.*: returns class and corresponding object id
def handle_object(obj, result={}):
    global OBJECTS

    obj_str = str(type(obj))

    if obj is not None:
        result['object'] = True
        result['class'] = obj.__class__.__name__

        class_str = str(obj.__class__)

        # if the resulting object can be pickled just send the raw object
        # REFACTOR:
        if 'types.Row' in obj_str or 'pandas.' in obj_str or list == type(obj):
            logger.info('Pickling object type %s' % (str(type(obj))))
            result['class'] = 'pickle'
            result['value'] = str(base64.b64encode(pickle.dumps(obj, 2)), 'utf-8')
        elif 'pyspark' in class_str or 'py4j' in class_str or type(obj) == types.FunctionType:
            id = str(uuid.uuid4())
            OBJECTS[id] = obj

            result['object'] = True
            result['object_id'] = id
            result['path'] = f'{obj.__module__}.{obj.__class__.__name__}'
            result['module'] = obj.__module__
            result['class'] = obj.__class__.__name__
        else:
            result['value'] = obj

    return result

def call_worker():
    global OBJECTS, REQ_QUEUE, RESP_QUEUE, RESPONSE_CACHE, Capture, logger

    while True:
        req = REQ_QUEUE.get()
        digest = req['digest']

        logger.info('Request:')
        logger.info(req)

        if digest in RESPONSE_CACHE:
            logger.info(f'Returning cached response for {digest}')

            resp = RESPONSE_CACHE[digest]
            resp['cached'] = True

            RESP_QUEUE.put(resp)
            continue

        res_obj = None

        resp = {
            'status': 'complete',
            'object': False,
            'object_id': None,
            'module': None,
            'class': None,
            'exception': None,
            'stdout': [],
            'cached': False,
            'digest': digest
        }

        if req['object_id'] != None:
            obj = OBJECTS[req['object_id']]

            if req['function'] != None:
                callable_obj = getattr(obj, req['function'])
            else:
                callable_obj = obj
        else:
            # have to loop through the module path and getattr
            # one by one because importing the whole module str
            # isn't always valid. For example pyspark.sql.session.SparkSession.Builder
            module_paths = req['path'].split('.')
            base_module = __import__(module_paths[0])

            module = base_module

            if len(module_paths) > 1:
                for m in module_paths[1:]:
                    module = getattr(module, m)

            if req['function'] != None:
                callable_obj = getattr(module, req['function'])
            # if no function is passed, it means the base object needs init
            else:
                callable_obj = module

        args, kwargs = parse_request_args(req['function'], req['args'], req['kwargs'])

        try:
            with Capture() as stdout:
                if req['is_property']:
                    res_obj = callable_obj
                elif req['is_item']:
                    res_obj = callable_obj[req['function']]
                else:
                    res_obj = callable_obj(*args, **kwargs)

                resp['stdout'] = stdout
        except Exception as e:
            resp['exception'] = str(e)

        resp = handle_object(res_obj, resp)

        logger.info('Response:')
        logger.info(resp)

        if req['cache'] and resp['exception'] == None:
            RESPONSE_CACHE[digest] = resp

        RESP_QUEUE.put(resp)
        REQ_QUEUE.task_done()

@app.before_request
def authenticate():
    if app.auth_token != None:
        auth = None

        if 'Auth' in request.headers:
            auth = request.headers['Auth']

        if auth != app.auth_token:
            return 'Authentication failed', 401


@app.route('/call', methods=['POST'])
def call():
    req = request.json
    digest = req['digest']

    if digest in RESPONSE_CACHE:
        logger.info(f'Returning cached response for {digest}')

        resp = RESPONSE_CACHE[digest]
        resp['cached'] = True
    else:
        resp = {'status': 'pending'}
        REQ_QUEUE.put(req)

    return jsonify(resp)

@app.route('/response', methods=['GET'])
def response():
    num = 0

    # pesudo long polling. give some time for the action
    # to run but dont keep open forever
    while RESP_QUEUE.empty():
        time.sleep(0.01+(num/100))
        num += 1

        if num > 100:
            break

    if RESP_QUEUE.empty():
        resp = {'status':'pending'}
    else:
        resp = RESP_QUEUE.get()
        RESP_QUEUE.task_done()

    return jsonify(resp)

@app.route('/health', methods=['GET'])
def health():
    return 'OK'

@app.route('/clear', methods=['POST', 'GET'])
def clear():
    global OBJECTS, RESPONSE_CACHE

    if len(OBJECTS) > 0:
        sc = pyspark.SparkContext.getOrCreate()
        sc.stop()

        OBJECTS = {}
        RESPONSE_CACHE = {}

    return 'OK'

def run(*args, **kwargs):
    threading.Thread(target=call_worker, daemon=True).start()

    if 'debug' not in kwargs or ('debug' in kwargs and kwargs['debug'] == False):
        app.logger.removeHandler(default_handler)
        app.logger = logger

        logger.info('Starting pysparkrpc web server')

    if 'auth' in kwargs and kwargs['auth'] != '':
        app.auth_token = kwargs['auth']
    else:
        app.auth_token = None

    del kwargs['auth']

    if 'port' not in kwargs:
        kwargs['port'] = 8765

    app.run(*args, **kwargs)

if __name__ == '__main__':
	configure_logging(True, 'INFO')
	run(host='127.0.0.1', debug=True, use_reloader=False, port=8765, auth='abc123')



import sys
import os
import time
import argparse

from pysparkrpc.server.logger import logger, configure_logging

PID_PATH = '/var/run/pysparkrpc.pid'

def start(args):
    if status(False):
        print('Pysparkrpc already running.')
    else:
        configure_logging(args.foreground, args.log_level)

        if args.foreground:
            start_server(args)
        else:
            start_background(args)

def start_background(args):
    try:
        pid = os.fork()

        if pid != 0:
            pid_file = open(PID_PATH, 'w')
            pid_file.write(str(pid))
            pid_file.close()
    except OSError as e:
        ## some debug output
        sys.exit(1)
    if pid == 0:
        start_server(args)

def start_server(args):
    import pysparkrpc.server as server
    server.run(host=args.host, port=args.port, auth=args.auth)

def load_pid():
    try:
        return int(open(PID_PATH).read())
    except IOError:
        return False

def status(print_status=True):
    pid = load_pid()

    if pid == False:
        if print_status:
            print('Pysparkrpc server is not running.')

        return False
    else:
        try:
            os.kill(pid, 0)
        except OSError:
            if print_status:
                print('Pysparkrpc server is not running.')

            return False
        else:
            if print_status:
                print('Pysparkrpc server is running. Pid: %d' % pid)

            return True

def stop():
    if status(False):
        pid = load_pid()
        os.kill(pid, 15)

        print('Pysparkrpc server stopped.')
    else:
        print('Pysparkrpc server is not running.')

def main():
    parser = argparse.ArgumentParser(description='Pysparkrpc Server')

    required = parser.add_argument_group('required arguments')
    required.add_argument('action',help='start|stop|status', choices=['start', 'stop', 'status'])

    parser.add_argument('--port', help='Port to bind to. Default: 8765', type=int, default=8765)
    parser.add_argument('--host', help='Host to bind to. Default: 0.0.0.0', type=str, default='0.0.0.0')
    parser.add_argument('--auth', help='Set an authentication token to authenticate the client', type=str, default='')
    parser.add_argument('--foreground', help='Starts server in the foreground', action='store_true')
    parser.add_argument('--log-level', help='Sets logging level. Default: INFO', type=str, default='INFO')

    args = parser.parse_args()

    if args.action == 'start':
        start(args)
    elif args.action == 'stop':
        stop()
    elif args.action == 'status':
        status()

if __name__ == "__main__":
    main()



import types
import functools
import base64
import pickle
import time
import hashlib
import json

import httpx
import cloudpickle

import pysparkrpc

PROXY_URL = 'http://localhost:8765'
CACHING = True

def _copy_func(f):
    g = types.FunctionType(f.__code__, f.__globals__, name=f.__name__,
                           argdefs=f.__defaults__,
                           closure=f.__closure__)
    g = functools.update_wrapper(g, f)
    return g

class APIClient(object):
    response_cache = {}
    http = httpx.Client(timeout=60.0)

    @classmethod
    def call(cls, object_id, path, function, args=(), kwargs={}, is_property=False, is_item=False, create=False):
        if function in pysparkrpc.PICKLE_FUNCS:
            function_args = [
               str(base64.b64encode(cloudpickle.dumps(args)), 'utf-8'),
               str(base64.b64encode(cloudpickle.dumps(kwargs)), 'utf-8'),
            ]
        else:
            function_args = cls._prepare_args(args, kwargs)

        req = {
            'object_id': object_id,
            'path': path,
            'function': function,
            'args': function_args[0],
            'kwargs': function_args[1],
            'is_property': is_property,
            'is_item': is_item,
            'cache': True
        }

        # if we want to force no caching then add a salt to
        # the payload to create a unique digest
        if CACHING == False:
            req['cache'] = False

        digest = hashlib.sha1(json.dumps(req).encode('utf-8')).hexdigest()
        req['digest'] = digest

        print(req)

        if digest in cls.response_cache and req['cache']:
            print(f'Using local cache for {digest}')
            resp = cls.response_cache[digest]
            resp['cached'] = True

            return cls._handle_response(resp, create)

        r = cls.http.post(PROXY_URL+'/call', json=req)
        resp = r.json()

        while True:
            if resp['status'] == 'complete' and resp['digest'] == digest:
                if resp['exception'] == None:
                    cls.response_cache[digest] = resp

                return cls._handle_response(resp, create)

            r = cls.http.get(PROXY_URL+'/response')
            resp = r.json()

    @classmethod
    def clear(cls):
        cls.http.get(PROXY_URL+'/clear')

    @classmethod
    def _handle_response(cls, resp, create):
        if resp['stdout'] != []:
            print('\n'.join(resp['stdout']))

        if resp['exception']:
            raise Exception(resp['exception'])

        if resp['object']:
            if resp['object_id'] != None:
                obj_id = resp['object_id']

                # udfs
                if resp['class'] == 'function':
                    f_code = compile(f'def proxyfunc(*args, **kwargs): return APIClient.call("{obj_id}", None, None, args, kwargs)', '<string>', 'exec')
                    f_func = types.FunctionType(f_code.co_consts[0], globals())

                    return f_func
                elif resp['class'] == 'JavaObject':
                    return pysparkrpc.proxy.ProxyJavaObject(_id=resp['object_id'], _cached=resp['cached'])
                else:
                    # don't initalize a new object if this is getting called from __init__
                    if create:
                        return obj_id
                    else:
                        return getattr(pysparkrpc, resp['class'])(_id=obj_id, _cached=resp['cached'])

            elif resp['class'] == 'pickle':
                return pickle.loads(base64.b64decode(resp['value']))
            else:
                return resp['value']
        else:
            return None

    @classmethod
    def _prepare_args(cls, args, kwargs):
        prepared_args = []
        prepared_kwargs = {}

        for a in args:
            arg_type = type(a)

            # pyspark objects can sometimes be in lists so we need to
            # check the list and send their id over so the server knows
            # what to retrieve
            if arg_type == list or arg_type == tuple:
                processed_list = []

                for x in a:
                    type_x = type(x)

                    if type_x == list or type_x == tuple:
                        processed_sub_list = []

                        for sub_x in x:
                            processed_sub_list.append(cls._proxy_obj_replace(sub_x))

                        processed_list.append(processed_sub_list)
                    else:
                        processed_list.append(cls._proxy_obj_replace(x))

                prepared_args.append(processed_list)
            elif arg_type == types.FunctionType:
                pickled_f = str(base64.b64encode(cloudpickle.dumps(_copy_func(a))), 'utf-8')
                prepared_args.append({'_CLOUDPICKLE': pickled_f})
            elif 'pyspark.sql.types' in str(arg_type):
                prepared_args.append({'_CLOUDPICKLE': str(base64.b64encode(cloudpickle.dumps(a)), 'utf-8')})
            else:
                prepared_args.append(cls._proxy_obj_replace(a))

        for a in kwargs:
            v = kwargs[a]
            prepared_kwargs[a] = cls._proxy_obj_replace(v)

        return prepared_args, prepared_kwargs

    @classmethod
    def _proxy_obj_replace(cls, obj):
        if hasattr(obj, '_PROXY'):
            return {'_PROXY_ID': obj._id}
        else:
            return obj



import os
import sys
import logging
from logging.handlers import TimedRotatingFileHandler

LOG_LEVELS = {
    'CRITICAL': logging.CRITICAL,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG
    }

LOG_PATH = '/var/log/pysparkrpc'

logger = None

def configure_logging(stdout, level):
    global logger
    formatter = logging.Formatter('%(asctime)s [%(levelname)s] - %(message)s')

    logger = logging.getLogger()
    logger.setLevel(LOG_LEVELS[level])

    if stdout:
        handler = logging.StreamHandler(stream=sys.stdout)
    else:
        if not os.path.exists(LOG_PATH):
            os.makedirs(LOG_PATH)

        handler = TimedRotatingFileHandler(
               LOG_PATH+'/server.log',
               when='midnight',
               interval=1,
               backupCount=7)

    handler.setFormatter(formatter)
    logger.addHandler(handler)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pysparkrpc', 'pysparkrpc.server']

package_data = \
{'': ['*']}

install_requires = \
['cloudpickle>=1.5.0,<2.0.0',
 'findspark>=1.4.2,<2.0.0',
 'flask>=1.1.2,<2.0.0',
 'httpx>=0.13.3,<0.14.0']

entry_points = \
{'console_scripts': ['pysparkrpc = pysparkrpc.server.cli:main']}

setup_kwargs = {
    'name': 'pysparkrpc',
    'version': '0.6.0',
    'description': 'Seamlessly run Pyspark code on remote clusters.',
    'long_description': 'PysparkRPC |teststatus|\n========================\n\n.. |teststatus| image:: https://github.com/abronte/PysparkRPC/workflows/Tests/badge.svg\n\n\nSeamlessly run Pyspark code on remote clusters.\n',
    'author': 'Adam Bronte',
    'author_email': 'adam@bronte.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/abronte/PysparkRPC',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)



__version__ = '0.6.0'

from types import FunctionType
import sys
import os

# findspark might not work for virtual env or non-standard
# install path but spark still may be available to python
try:
    import findspark
    findspark.init()
except:
    pass

import httpx
import pyspark
from pyspark.version import __version__ as PYSPARK_VERSION
import pyspark.sql.functions
import pyspark.ml

from pysparkrpc.proxy import Proxy, ProxyJavaObject
import pysparkrpc.api_client as api
from pysparkrpc.api_client import APIClient

TARGET_OBJS = [
    pyspark.rdd.RDD,
    pyspark.rdd.PipelinedRDD,
    pyspark.context.SparkContext,
    pyspark.sql.context.SQLContext,
    pyspark.sql.DataFrame,
    pyspark.sql.DataFrameReader,
    pyspark.sql.DataFrameWriter,
    pyspark.sql.group.GroupedData,
    pyspark.sql.session.SparkSession,
    pyspark.sql.column.Column,
    pyspark.sql.udf.UDFRegistration
]

# pyspark will make classes available via the root module
# we need to specifically inject our clases into these as well
TARGET_ALIAS = {
    'SparkContext': 'pyspark',
    'SQLContext': 'pyspark.sql',
    'SparkSession': 'pyspark.sql'
}

TARGET_ALL_MODULES = [
    pyspark.ml.classification,
    pyspark.ml.clustering,
    pyspark.ml.regression,
    pyspark.ml.fpm,
    pyspark.ml.recommendation
]

MONKEY_PATCH_FUNCS = {
    pyspark.rdd.RDD: ['toDF']
}

MONKEY_PATCH_PROPS = {
    pyspark.context.SparkContext: ['_jsc']
}

for m in TARGET_ALL_MODULES:
    for c in getattr(m, '__all__'):
        TARGET_OBJS.append(getattr(m, c))

TARGET_FUNCTIONS = [
    pyspark.sql.functions
]

BUILTIN_FUNCS = [
    '__eq__', '__ne__', '__lt__', '__le__', '__ge__', '__gt__',
    '__neg__', '__add__', '__radd__', '__rsub__', '__mul__', '__rmul__',
    '__div__', '__rdiv__', '__rtruediv__', '__rmod__', '__pow__', '__rpow__',
    '__and__', '__or__', '__invert__', '__rand__', '__ror__',
    '__getattr__', '__getitem__', '__repr__'
]

PICKLE_FUNCS = [
    'parallelize',
    'createDataFrame',
    'predict'
]

def inject(url='http://localhost:8765', auth=None):
    if auth != None:
        APIClient.http = httpx.Client(timeout=60.0, headers={'Auth':auth})

    api.PROXY_URL=url

    _handle_spark_versions()

    for m in TARGET_FUNCTIONS:
        _build_functions(m)

    for c in TARGET_OBJS:
        obj_name = c.__name__
        proxy_class = type(obj_name, (Proxy,), {})
        globals()[obj_name] = proxy_class

        _build_class(c)

    for c in TARGET_OBJS:
        obj_name = c.__name__
        obj_module = c.__module__

        setattr(sys.modules[obj_module], obj_name, globals()[obj_name])

        if obj_name in TARGET_ALIAS:
            setattr(sys.modules[TARGET_ALIAS[obj_name]], obj_name, globals()[obj_name])

def cache(set_to=True):
    api.CACHING = set_to

    if set_to:
        print('PysparkRPC caching ON')
    else:
        print('PysparkRPC caching OFF')

def clear():
    APIClient.clear()

def _handle_spark_versions():
    major = PYSPARK_VERSION[0]

    if major == '3':
        import pyspark.ml.functions
        TARGET_FUNCTIONS.append(pyspark.ml.functions)

def _build_class(obj, path=None):
    obj_name = obj.__name__
    obj_module = obj.__module__

    # don't patch any non-pyspark objects
    if 'pyspark' not in obj_module:
        return

    if path == None:
        path = f'{obj.__module__}.{obj_name}'

    proxy_class = globals()[obj_name]
    setattr(proxy_class, '_path', path)

    obj_names = dir(obj)

    for f in BUILTIN_FUNCS:
        # don't override the builtin if pyspark didn't override it
        # built in functions return as the class wrapper_descriptor
        if f in obj_names and getattr(obj, f).__class__.__name__ != 'wrapper_descriptor':
            patch_code = compile(f'def {f}(self, *args, **kwargs): return APIClient.call(self._id, self._path, "{f}", args, kwargs)', '<string>', 'exec')
            patch_func = FunctionType(patch_code.co_consts[0], globals())
            setattr(proxy_class, f, patch_func)

    if obj in MONKEY_PATCH_FUNCS:
        for f in MONKEY_PATCH_FUNCS[obj]:
            patch_code = compile(f'def {f}(self, *args, **kwargs): return APIClient.call(self._id, self._path, "{f}", args, kwargs)', '<string>', 'exec')
            patch_func = FunctionType(patch_code.co_consts[0], globals())
            setattr(proxy_class, f, patch_func)

    if obj in MONKEY_PATCH_PROPS:
        for f in MONKEY_PATCH_PROPS[obj]:
            patch_code = compile(f'def {f}(self): return APIClient.call(self._id, self._path, "{f}", is_property=True)', '<string>', 'exec')
            patch_func = FunctionType(patch_code.co_consts[0], globals())
            setattr(proxy_class, f, property(patch_func))

    for f in obj_names:

        if ('__' in f and f not in BUILTIN_FUNCS) or f[0] == '_':
            continue

        class_name = getattr(obj, f).__class__.__name__

        if hasattr(getattr(obj, f), '__self__'):
            patch_code = compile(f'def {f}(cls, *args, **kwargs): return APIClient.call(None, cls._path, "{f}", args, kwargs)', '<string>', 'exec')
            patch_func = FunctionType(patch_code.co_consts[0], globals())

            setattr(proxy_class, f, classmethod(patch_func))
        elif class_name == 'property':
            patch_code = compile(f'def {f}(self, *args, **kwargs): return APIClient.call(self._id, self._path, "{f}", is_property=True)', '<string>', 'exec')
            patch_func = FunctionType(patch_code.co_consts[0], globals())

            setattr(proxy_class, f, property(patch_func))
        elif class_name == 'function':
            patch_code = compile(f'def {f}(self, *args, **kwargs): return APIClient.call(self._id, self._path, "{f}", args, kwargs)', '<string>', 'exec')
            patch_func = FunctionType(patch_code.co_consts[0], globals())

            setattr(proxy_class, f, patch_func)
        elif class_name != 'function' and class_name != 'type':
            prop_class = getattr(obj, f).__class__
            prop_class_name = prop_class.__name__

            c = type(prop_class_name, (Proxy,), {'_propclass':True})
            globals()[prop_class_name] = c

            _build_class(prop_class, path=f'{obj_module}.{obj_name}.{f}')

            # assuming the class has to be initialized
            # TODO: maybe this should be added to the base class init?
            setattr(proxy_class, f, c())

def _build_functions(module):
    for f in dir(module):
        class_name = getattr(module, f).__class__.__name__

        if '__' not in f and f[0] != '_' and class_name == 'function':
            patch_code = compile(f'def {f}(*args, **kwargs): return APIClient.call(None, "{module.__name__}", "{f}", args, kwargs)', '<string>', 'exec')
            patch_func = FunctionType(patch_code.co_consts[0], globals())
            setattr(module, f, patch_func)



from pysparkrpc.api_client import APIClient

import cloudpickle
import sys

class Proxy(object):
    _PROXY = True
    _path = None
    _propclass = False
    _cached = False

    def __init__(self, *args, **kwargs):
        self._kwargs = kwargs
        self._class = self.__class__.__name__
        self._args = args
        self._id = None

        if '_cached' in kwargs:
            self._cached = kwargs['_cached']

        if '_id' in kwargs:
            self._id = kwargs['_id']
        elif self._path and self._propclass == False:
            self._id = APIClient.call(None, self._path, None, args, kwargs, create=True)

class ProxyJavaObject(Proxy):
    def __getattr__(self, name):
        def method(*args, **kwargs):
            return APIClient.call(self._id, None, name, args, kwargs)

        return method
